from datasets import Dataset
import requests
import os
import tempfile
import uuid


def get_presigned_url(file_name: str, bucket_name: str = "creaodev"):
    url = "https://drk3mkqa6b.execute-api.us-west-2.amazonaws.com/default/minimaxserver"
    api_key = os.environ["CREAO_API_KEY"]
    headers = {"Content-Type": "application/json", "x-api-key": api_key}
    # Payload to be sent to the Lambda function via API Gateway
    payload = {"action": "s3", "file_name": file_name, "bucket_name": bucket_name}
    # Send the request
    response = requests.post(url, headers=headers, json=payload)
    print("reponse:", response.json())
    return response.json()["url"]


def submit_pipeline_job(
    pipeline_name: str, pipe_yaml: str, run_config_template: str, dataset: Dataset
):
    """
    Submit a pipeline job to the Creao server
    :param pipeline_name: The name of the pipeline
    :param pipe_yaml: The pipeline YAML
    :param run_config_template: The run configuration template
    :param dataset: The dataset to use
    :return: The response from the server
    """
    pipeline_id = str(uuid.uuid4())
    pipe_yaml = pipe_yaml.replace("pipeline_id_default", pipeline_id)
    # generate a random file name for the dataset based on uuid
    file_name = str(uuid.uuid4()) + ".parquet"
    bucket_name = "creaodev"
    temp_file_path = f"/tmp/{file_name}"
    # save the dataset to a temporary file
    # Use tempfile to create a temporary file
    # Save the dataset as a Parquet file
    dataset.to_parquet(temp_file_path)
    print(f"Dataset saved to temporary file: {temp_file_path}")

    # get the presigned url for the dataset
    presigned_url = get_presigned_url(file_name, bucket_name)

    # Open the zip file and upload it using the pre-signed URL
    with open(temp_file_path, "rb") as f:
        upload_response = requests.put(presigned_url, data=f)

    # Check the response from the upload
    if upload_response.status_code == 200:
        print("File uploaded successfully")
    else:
        print(
            "Upload failed with status code:",
            upload_response.status_code,
            upload_response.text,
        )
    print("final clean up")
    # (Optional) Clean up the temporary file when done
    if os.path.exists(temp_file_path):
        os.remove(temp_file_path)
        print(f"Temporary file {temp_file_path} has been removed.")

    # generate a uuid
    s3_path = f"s3://{bucket_name}/{file_name}"
    api_key = os.environ["CREAO_API_KEY"]
    payload = {
        "pipeline_id": pipeline_id,
        "pipeline_name": pipeline_name,
        "pipe_yaml": pipe_yaml,
        "run_config_template": run_config_template,
        "dataset_s3_path": s3_path,
        "api_key": api_key,
        "action": "pipeline",
    }
    # The API Gateway endpoint URL
    url = "https://drk3mkqa6b.execute-api.us-west-2.amazonaws.com/default/minimaxserver"
    headers = {"Content-Type": "application/json", "x-api-key": api_key}
    # Payload to be sent to the Lambda function via API Gateway
    # Send the request
    response = requests.post(url, headers=headers, json=payload)
    # print("reponse:", response.content)
    try:
        return response.json()
    except Exception as e:
        print(f"create pipeline failed with error:{e}, with response:{response.text}")



from typing import List
from haystack import component, Pipeline

import pandas as pd
from datasets import Dataset, concatenate_datasets

from github import Github, Repository, ContentFile
import requests
import os

from typing import List, Optional
from pathlib import Path

import nltk
# nltk.download('punkt_tab')
# nltk.download('averaged_perceptron_tagger_eng')


def download(c: ContentFile, out: str) -> None:
    r = requests.get(c.download_url)
    output_path = f"{out}/{c.path}"
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    with open(output_path, "wb") as f:
        print(f"downloading {c.path} to {out}")
        f.write(r.content)


def download_folder(repo: Repository, folder: str, out: str, recursive: bool) -> None:
    contents = repo.get_contents(folder)
    for c in contents:
        if c.download_url is None:
            if recursive:
                download_folder(repo, c.path, out, recursive)
            continue
        download(c, out)


def create_chunks(md_files: List[str]) -> dict[str, List[str]]:
    """Create the chunks of text from the markdown files.

    Note:
        We should allow the chunking strategy to take into account the max size delimited
        by the number of tokens.

    Args:
        md_files: List of paths to the markdown files.

    Returns:
        Dictionary from filename to the list of chunks.
    """
    from unstructured.chunking.title import chunk_by_title
    from unstructured.partition.auto import partition

    data = {}
    for file in md_files:
        partitioned_file = partition(filename=file)
        chunks = [str(chunk) for chunk in chunk_by_title(partitioned_file)]

        data[str(file)] = chunks
    return data


def create_dataset(
    data: dict[str, List[str]], repo_name: Optional[str] = None
) -> Dataset:
    """Creates a dataset from the dictionary of chunks.

    Args:
        data: Dictionary from filename to the list of chunks,
            as obtained from `create_chunks`.

    Returns:
        Dataset with `filename` and `chunks` columns.
    """
    df = pd.DataFrame.from_records(
        [(k, v) for k, values in data.items() for v in values],
        columns=["filename", "chunks"],
    )
    if repo_name:
        df["repo_name"] = repo_name
    ds = Dataset.from_pandas(df)
    return ds


@component
class RepoMarkDownCollector:
    """
    A component collecting MarkDown files from a github repo
    """

    @component.output_types(md_file_paths=List[str])
    def run(self, repo_name: str):
        print()
        print("[RepoMarkDownCollector]")
        gh = Github()
        repo = gh.get_repo(repo_name)

        docs_path = Path(repo_name.split("/")[1])  # the default output path

        if docs_path.exists():
            print(f"Folder {docs_path} already exists, skipping download.")
        else:
            print("Downloading the files...")
            download_folder(
                repo, "docs", str(docs_path), True
            )  # docs could be exposable

        print("Start collecting MarkDown files...")
        md_files = list(docs_path.glob("**/*.md"))  # a list of Path
        md_files = [str(path) for path in md_files]  # a list of str
        print(md_files)
        print(f"{len(md_files)} MarkDown files collected successfully")
        print("[RepoMarkDownCollector]")
        print()
        return {"md_file_paths": md_files}


@component
class ChunkDataGenerator:
    """
    A component generating chunks from a list of files
    """

    @component.output_types(chunk_data=dict[str, List[str]])
    def run(self, source_files: List[str]):
        # Loop to iterate over the files and generate chunks from the text pieces
        print()
        print("[ChunkDataGenerator]")
        print("Generating the chunks from the markdown files...")
        output = create_chunks(source_files)
        print("Completed.")
        print("[ChunkDataGenerator]")
        print()
        return {"chunk_data": output}


@component
class DatasetCreator:
    """
    A component generating Dataset with filename and chunks columns and repo name
    """

    @component.output_types(dataset=Dataset)
    def run(self, chuck_data: dict[str, List[str]], repo_name: Optional[str] = None):
        print()
        print("[DatasetCreator]")
        print("Creating dataset...")
        output = create_dataset(chuck_data, repo_name=repo_name)
        print("Completed.")
        print("[DatasetCreator]")
        print()
        return {"dataset": output}


data_pipeline = Pipeline()
data_pipeline.add_component(name="markdown_collector", instance=RepoMarkDownCollector())
data_pipeline.add_component(name="chunk_data_generator", instance=ChunkDataGenerator())
data_pipeline.add_component(name="dataset_creator", instance=DatasetCreator())

data_pipeline.connect(
    sender="markdown_collector.md_file_paths",
    receiver="chunk_data_generator.source_files",
)
data_pipeline.connect(
    sender="chunk_data_generator.chunk_data", receiver="dataset_creator.chuck_data"
)

result = data_pipeline.run(
    {"markdown_collector": {"repo_name": "argilla-io/argilla-python"}}
)

print(result["dataset_creator"]["dataset"])



from typing import Any, Dict, List
from haystack import component, default_from_dict, default_to_dict
from openai import OpenAI
from haystack.utils import (
    Secret,
    deserialize_callable,
    deserialize_secrets_inplace,
    serialize_callable,
)


@component
class Embed:
    """
    Embed text using a pre-trained model from NVIDIA.

    ### Usage example
    ```python
        from component.generators.embed import Embed
        embed = Embed(service="NVIDIA")
        response = embed.run("how are you")
    ```
    """

    def __init__(
        self,
        service: str,
        encoding_format: str = "float",
    ):
        """
        Initialize the component.
        :param service: The service to use for embedding. Currently only "NVIDIA" is supported.
        :param encoding_format: The encoding format to use. Currently only "float" is supported.

        """
        api_key = ""
        base_url = ""
        self.encoding_format = encoding_format
        self.service = service
        if service == "NVIDIA":
            api_key = Secret.from_env_var("BUILD_NVIDIA_API_KEY")
            base_url = "https://ai.api.nvidia.com/v1/retrieval/nvidia"
            self.model = "NV-Embed-QA"
        else:
            raise ValueError(f"Unsupported service '{service}'")
        self.client = OpenAI(
            base_url=base_url,
            api_key=api_key.resolve_value(),
        )

    @component.output_types(embeddings=List[float])
    def run(self, text: str):
        embeddings = (
            self.client.embeddings.create(
                input=[text],
                model=self.model,
                encoding_format=self.encoding_format,
                extra_body={"input_type": "query", "truncate": "NONE"},
            )
            .data[0]
            .embedding
        )
        return {"embeddings": embeddings}

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize this component to a dictionary.

        :returns:
            The serialized component as a dictionary.
        """
        return default_to_dict(
            self,
            service=self.service,
            encoding_format=self.encoding_format,
        )

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Embed":
        """
        Deserialize this component from a dictionary.

        :param data:
            The dictionary representation of this component.
        :returns:
            The deserialized component instance.
        """
        return default_from_dict(cls, data)



from typing import Any, Dict, List
from creao.core.Endpoints import CreaoLLM
from haystack import component, default_from_dict, default_to_dict
import json


@component
class CreaoGenerator:
    def __init__(self):
        self.llm = CreaoLLM()

    @component.output_types(replies=List[str], meta=List[Dict[str, Any]])
    def run(self, prompt: str):
        response_raw = self.llm.invoke(prompt, "")
        i = 0
        while i < 3:
            if "reply" in response_raw:
                break
            response_raw = self.llm.invoke(prompt, "")
            i += 1
        if "reply" not in response_raw:
            return {"replies": [""], "meta": [{"service": "minimax"}]}
        return {"replies": [response_raw["reply"]], "meta": [{"service": "minimax"}]}

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize this component to a dictionary.

        :returns:
            The serialized component as a dictionary.
        """
        return default_to_dict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "CreaoGenerator":
        """
        Deserialize this component from a dictionary.

        :param data:
            The dictionary representation of this component.
        :returns:
            The deserialized component instance.
        """
        return default_from_dict(cls, data)



from typing import Any, Dict, List
from haystack import component, default_from_dict, default_to_dict, Document
from creao.core.Generator import *
import concurrent.futures
from creao.core.Endpoints import CreaoLLM


@component
class MapQuestionTypeToQuestion:
    """
    A component mapping question type to question
    """

    def __init__(
        self,
        type_of_questions: Dict[str, str],
        service: str = "default",
        pipeline_id: str = "pipeline_id_default",
    ):
        self.service = service
        self.pipeline_id = pipeline_id
        self.type_of_questions = type_of_questions
        if self.service == "default":
            self.llm = CreaoLLM(
                bot_name="question generation assistant",
                bot_content="You are given a list of interests and a passage. Your task is to generate questions that align with the interests and the passage. Reflective Thinking: Before you generate the final list of questions, pause to examine your assumptions, biases, and the mental models that the interests might use. Consider how the interests influence what they find interesting, and how they prioritize topics. Be open to learning from previous similar tasks and improving the outcome. Explicitly share your reflective thought process before giving the final answer.",
            )
        elif self.service == "openai":
            self.generator = Generator()

    @component.output_types(documents=List[Document])
    def run(self, question_mapping: List[Dict[str, str]], chunk: str, file_name: str):
        """
        Map question type to question
        :param question_type: The question type to map
        :return: The question
        """

        # print("generating question with minimax")
        def process_question_type(item):
            interest_questions = []
            for q_type in item["q_type"]:
                if q_type.lower() in self.type_of_questions:
                    if self.service == "default":
                        prompt = extract_questions_prompt.format(
                            interest=item["interest"],
                            types=self.type_of_questions[q_type.lower()],
                            file_name=file_name,
                            passage=chunk,
                        )
                        json_schema = {
                            "generated_questions": {
                                "type": "array",
                                "items": {"type": "string"},
                            }
                        }
                        response_json = self.llm.invoke(
                            prompt,
                            json_schema,
                            "MapQuestionTypeToQuestion",
                            self.pipeline_id,
                        )

                        try:
                            raw_answer = json.loads(response_json["reply"])
                        except Exception as e:
                            print(
                                f"MapQuestionTypeToQuestion json decode error:{e}, with response_json:{response_json}"
                            )
                            raw_answer = {"generated_questions": []}
                        if "generated_questions" in raw_answer:
                            questions = raw_answer["generated_questions"]
                        else:
                            questions = []
                    elif self.service == "openai":
                        questions = self.generator.generate_questions(
                            file_name,
                            chunk,
                            item["interest"],
                            self.type_of_questions[q_type.lower()],
                        )
                    interest_questions.extend(questions)
            return interest_questions

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            raw_questions = [
                question
                for result in executor.map(process_question_type, question_mapping)
                for question in result
            ]
            questions = [item for item in raw_questions if item != []]
        documents = []
        for question in questions:
            doc = Document(
                content=question, meta={"file_name": file_name, "chunk": chunk}
            )
            documents.append(doc)
        # print("generating question with minimax done")
        return {"documents": documents}

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize this component to a dictionary.

        :returns:
            The serialized component as a dictionary.
        """
        return default_to_dict(
            self,
            type_of_questions=self.type_of_questions,
            service=self.service,
            pipeline_id=self.pipeline_id,
        )

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "MapQuestionTypeToQuestion":
        """
        Deserialize this component from a dictionary.

        :param data:
            The dictionary representation of this component.
        :returns:
            The deserialized component instance.
        """
        return default_from_dict(cls, data)



from typing import List
from haystack import component, Pipeline

import torch
from sentence_transformers import SentenceTransformer
from sentence_transformers.evaluation import (
    InformationRetrievalEvaluator,
    SequentialEvaluator,
)
from sentence_transformers.util import cos_sim

from datasets import Dataset
from sentence_transformers import SentenceTransformerModelCardData, SentenceTransformer
import time
from datetime import datetime
from sentence_transformers.losses import MatryoshkaLoss, TripletLoss

from sentence_transformers import SentenceTransformerTrainingArguments
from sentence_transformers.training_args import BatchSamplers

from sentence_transformers import SentenceTransformerTrainer
from datasets import load_dataset, concatenate_datasets

from component.evaluator import prepare_matryoshka_evaluator


def load_split_dataset_to_local(
    dataset_path: str,
    select_columns: List[str],
    test_ratio: float,
    train_local_json_path: str,
    test_local_json_path: str,
) -> dict[str, str]:
    """Load dataset, split and save to local.

    Args:
        xxx

    Returns:
        Two local Json file paths organized in a dict, representing the split train and test.
    """
    # dataset_path = "plaguss/argilla_sdk_docs_queries"
    # select_columns = ["anchor", "positive", "negative"]

    # Load dataset from the hub
    dataset = load_dataset(dataset_path, split="train")
    dataset = (
        load_dataset(dataset_path, split="train")
        .select_columns(select_columns)  # Select the relevant columns
        .add_column("id", range(len(dataset)))  # Add an id column to the dataset
        .train_test_split(
            test_size=test_ratio
        )  # split dataset by specifying test set rate
    )

    # save datasets to disk
    dataset["train"].to_json(train_local_json_path, orient="records")
    dataset["test"].to_json(test_local_json_path, orient="records")
    return {"train": train_local_json_path, "test": test_local_json_path}


@component
class HuggfingFaceDatasetLoadSplitPersistence:
    """
    A component loading, spliting, saving datasets
    """

    @component.output_types(dict_train_test_paths=dict[str, str])
    def run(
        self,
        dataset_path: str,
        select_columns: List[str],
        test_ratio: float,
        train_local_json_path: str,
        test_local_json_path: str,
    ):
        print()
        output_dict = load_split_dataset_to_local(
            dataset_path,
            select_columns,
            test_ratio,
            train_local_json_path,
            test_local_json_path,
        )
        print()
        return {"dict_train_test_paths": output_dict}


"""
test = HuggfingFaceDatasetLoadSplitPersistence()
result = test.run(dataset_path="plaguss/argilla_sdk_docs_queries", select_columns=["anchor", "positive", "negative"], test_ratio=0.1, 
         train_local_json_path="train_dataset.json", test_local_json_path="test_dataset.json")
print(result["dict_train_test_paths"])
"""


finetuner_config_example = {
    "train_data_path": "train_dataset.json",
    "test_data_path": "test_dataset.json",
    "select_columns": ["anchor", "positive", "negative"],
    "model_id": "BAAI/bge-base-en-v1.5",
    "matryoshka_dimensions": [768, 512, 256, 128, 64],
    # "device": "mps",
    "output_dir": "bge-base-argilla-sdk-matryoshka",
    "num_train_epochs": 8,  # number of epochs
    "per_device_train_batch_size": 16,  # train batch size
    "gradient_accumulation_steps": 8,  # for a global batch size of 512
    "per_device_eval_batch_size": 4,  # evaluation batch size
    "warmup_ratio": 0.1,  # warmup ratio
    "learning_rate": 2e-7,  # learning rate, 2e-5 is a good value
    "lr_scheduler_type": "cosine",  # use constant learning rate scheduler
    # optim="adamw_torch_fused",                  # use fused adamw optimizer
    # tf32=True,                                  # use tf32 precision
    # bf16=True,                                  # use bf16 precision
    # batch_sampler=BatchSamplers.NO_DUPLICATES,  # MultipleNegativesRankingLoss benefits from no duplicate samples in a batch
    "eval_strategy": "epoch",  # evaluate after each epoch
    "save_strategy": "epoch",  # save after each epoch
    "logging_steps": 5,  # log every 10 steps
    "save_total_limit": 1,  # save only the last 3 models
    "load_best_model_at_end": True,  # load the best model when training ends
    "metric_for_best_model": "eval_dim_512_cosine_ndcg@10",  # Optimizing for the best ndcg@10 score for the 512 dimension
}


def doEmbederMatryoshkaTripletFinetune(finetuner_config: dict) -> str:
    # load dataset back
    test_dataset = load_dataset(
        "json", data_files=finetuner_config["test_data_path"], split="train"
    )
    train_dataset = load_dataset(
        "json", data_files=finetuner_config["train_data_path"], split="train"
    )
    # corpus_dataset = concatenate_datasets([train_dataset, test_dataset]) #only needed for evaluation

    # remove none
    train_dataset_cleaned = (
        train_dataset.select_columns(finetuner_config["select_columns"])
        .to_pandas()
        .dropna()
    )
    test_dataset_cleaned = (
        test_dataset.select_columns(finetuner_config["select_columns"])
        .to_pandas()
        .dropna()
    )

    train_dataset_cleaned = Dataset.from_pandas(
        train_dataset_cleaned, preserve_index=False
    )
    test_dataset_cleaned = Dataset.from_pandas(
        test_dataset_cleaned, preserve_index=False
    )

    model_id = finetuner_config[
        "model_id"
    ]  # "BAAI/bge-base-en-v1.5"  # Hugging Face model ID https://huggingface.co/BAAI/bge-base-en-v1.5
    matryoshka_dimensions = finetuner_config[
        "matryoshka_dimensions"
    ]  # [768, 512, 256, 128, 64]  # Important: large to small

    device = (
        "mps"
        if torch.backends.mps.is_available()
        else "cuda"
        if torch.cuda.is_available()
        else "cpu"
    )
    print("device", device)

    # load model
    start_time = time.time()
    # load model with SDPA for using Flash Attention 2
    model = SentenceTransformer(
        model_id,
        # model_kwargs={"attn_implementation": "sdpa"},  # sdpa will be used by default if available
        model_card_data=SentenceTransformerModelCardData(
            language="en",
            license="apache-2.0",
            model_name="BGE base ArgillaSDK Matryoshka",
        ),
    )

    elapsed_time = time.time() - start_time
    print("model loading time", elapsed_time)

    # Define the loss
    inner_train_loss = TripletLoss(model)
    train_loss = MatryoshkaLoss(
        model, inner_train_loss, matryoshka_dims=matryoshka_dimensions
    )

    # define training arguments
    args = SentenceTransformerTrainingArguments(
        output_dir=finetuner_config[
            "output_dir"
        ],  # "bge-base-argilla-sdk-matryoshka", # output directory and hugging face model ID
        num_train_epochs=finetuner_config["num_train_epochs"],  # 3,  # number of epochs
        per_device_train_batch_size=finetuner_config[
            "per_device_train_batch_size"
        ],  # 8,             # train batch size
        gradient_accumulation_steps=finetuner_config[
            "gradient_accumulation_steps"
        ],  # 4,             # for a global batch size of 512
        per_device_eval_batch_size=finetuner_config[
            "per_device_eval_batch_size"
        ],  # 4,              # evaluation batch size
        warmup_ratio=finetuner_config[
            "warmup_ratio"
        ],  # 0.1,                           # warmup ratio
        learning_rate=finetuner_config[
            "learning_rate"
        ],  # 2e-5,                         # learning rate, 2e-5 is a good value
        lr_scheduler_type=finetuner_config[
            "lr_scheduler_type"
        ],  # "cosine",                 # use constant learning rate scheduler
        # NOTE: In colab we can work with the optimizer at least, but neither tf32 nor bf16
        #    optim="adamw_torch_fused",                  # use fused adamw optimizer
        #    tf32=True,                                  # use tf32 precision
        #    bf16=True,                                  # use bf16 precision
        # batch_sampler=BatchSamplers.NO_DUPLICATES,  # MultipleNegativesRankingLoss benefits from no duplicate samples in a batch
        eval_strategy=finetuner_config[
            "eval_strategy"
        ],  # "epoch",                      # evaluate after each epoch
        save_strategy=finetuner_config[
            "save_strategy"
        ],  # "epoch",                      # save after each epoch
        logging_steps=finetuner_config[
            "logging_steps"
        ],  # 5,                            # log every 10 steps
        save_total_limit=finetuner_config[
            "save_total_limit"
        ],  # 1,                         # save only the last 3 models
        load_best_model_at_end=finetuner_config[
            "load_best_model_at_end"
        ],  # True,                # load the best model when training ends
        metric_for_best_model=finetuner_config[
            "metric_for_best_model"
        ],  # "eval_dim_512_cosine_ndcg@10",  # Optimizing for the best ndcg@10 score for the 512 dimension
    )

    # prepare trainer (requiring evaluator)
    evaluator = prepare_matryoshka_evaluator(
        train_dataset, test_dataset, matryoshka_dimensions
    )
    trainer = SentenceTransformerTrainer(
        model=model,  # bg-base-en-v1
        args=args,  # training arguments
        train_dataset=train_dataset.select_columns(
            finetuner_config["select_columns"]
        ),  # training dataset
        loss=train_loss,
        evaluator=evaluator,
    )

    # start training, the model will be automatically saved to the hub and the output directory
    start_time = time.time()
    trainer.train()
    elapsed_time = time.time() - start_time
    print("model tuning elapsed time", elapsed_time)

    # save the best model
    trainer.save_model(finetuner_config["output_dir"])
    print("best finetuned model saved to:", finetuner_config["output_dir"])

    # push model to hub
    # trainer.model.push_to_hub("bge-base-argilla-sdk-matryoshka")
    return finetuner_config["output_dir"]


"""
print('start')
doEmbederMatryoshkaTripletFinetune(finetuner_config)
print('success')
"""


@component
class EmbeddingModelMatryoshkaTripletFinetuner:
    """
    A component loading data and model, finetuning with MatryoshkaLoss, and saving the best model, and optinally push to huggingface.
    """

    @component.output_types(model_output_path=str)
    def run(self, embed_finetuner_config: dict):
        print("[EmbeddingModelMatryoshkaTripletFinetuner]")
        print("started")
        model_output_path = doEmbederMatryoshkaTripletFinetune(
            finetuner_config=embed_finetuner_config
        )
        print("completed")
        print("[EmbeddingModelMatryoshkaTripletFinetuner]")
        return {"model_output_path": model_output_path}


finetunertest = EmbeddingModelMatryoshkaTripletFinetuner()
result = finetunertest.run(finetuner_config_example)
print(result["model_output_path"])



from openai import OpenAI
import time
import requests
import os


class Embed:
    def __init__(self):
        self.client = OpenAI()

    def invoke(self, text):
        response = self.client.embeddings.create(
            model="text-embedding-3-small",
            input=text,
        )
        return response.data[0].embedding


class CreaoLLM:
    def __init__(self, bot_name="assistant", bot_content="assistant") -> None:
        self.bot_name = bot_name
        self.bot_content = bot_content

    def invoke(self, prompt, schema, component_id="default", pipeline_id="default"):
        """
        Invoke the Creao LLM API
        """
        start_time = time.time()
        # The API Gateway endpoint URL
        url = "https://drk3mkqa6b.execute-api.us-west-2.amazonaws.com/default/minimaxserver"
        api_key = os.environ["CREAO_API_KEY"]
        headers = {"Content-Type": "application/json", "x-api-key": api_key}
        # Payload to be sent to the Lambda function via API Gateway
        payload = {
            "prompt": prompt,
            "schema": schema,
            "bot_name": self.bot_name,
            "bot_content": self.bot_content,
            "component_id": component_id,
            "pipeline_id": pipeline_id,
        }
        # Send the request
        response = requests.post(url, headers=headers, json=payload)
        response_josn = response.json()
        end_time = time.time()
        elapsed_time = end_time - start_time
        print("response_josn:", response_josn)
        print(f"CreaoLLM API call took total {elapsed_time} seconds")
        if "elapsed_time" in response_josn:
            minimax_elapsed_time = response_josn["elapsed_time"]
            print(f"Minimax server took {minimax_elapsed_time} seconds")
        # print("reponse:", response.content)
        try:
            return response_josn
        except Exception as e:
            print(f"CreaoLLM json decode error:{e}, with response:{response.text}")
            return {"reply": ""}


class OpenAILLM:
    def __init__(self, model_id="gpt-4o-mini"):
        self.client = OpenAI()
        self.model_id = model_id

    def invoke(self, prompt, schema=None):
        messages = [{"role": "user", "content": prompt}]
        try:
            if schema is None:
                response = self.client.chat.completions.create(
                    model=self.model_id,
                    messages=messages,
                    temperature=0,
                    max_tokens=1024,
                )
                return response.choices[0].message.content
            else:
                response = self.client.beta.chat.completions.parse(
                    model=self.model_id,
                    messages=messages,
                    temperature=0,
                    max_tokens=1024,
                    response_format=schema,
                )
            return response.choices[0].message.content
        except Exception as e:
            print(e)
            return None


# creaoLM = CreaoLLM()
# r = creaoLM.invoke("hello", {"name":"world"})
# print(r)



from typing import List
from haystack import component, default_from_dict, default_to_dict, Document
from creao.core.Generator import *
import concurrent.futures


@component
class IntelligentFilter:
    def __init__(
        self, service: str = "default", pipeline_id: str = "pipeline_id_default"
    ):
        self.service = service
        self.pipeline_id = pipeline_id
        if self.service == "default":
            self.llm = CreaoLLM(
                bot_name="question intelligence filter assistant",
                bot_content="You are given a question. Your task is to determine the intelligence of the question. Reflective Thinking: Before you generate the final decision, pause to examine your assumptions, biases, and the mental models that the question might use. Consider how the question influences the intelligence of the question, and how they prioritize the intelligence. Be open to learning from previous similar tasks and improving the outcome. Explicitly share your reflective thought process before giving the final answer.",
            )
        elif self.service == "openai":
            self.intelligent_question_filter = Intelligent_Question_Filter()

    @component.output_types(documents=List[Document])
    def run(self, documents: List[Document]):
        # print("filtering intelligence with minimax")
        if len(documents) == 0:
            return {"outputs": []}
        inputs = [doc.content for doc in documents]
        file_name = documents[0].meta["file_name"]
        chunk = documents[0].meta["chunk"]

        def filter_question(question):
            if self.service == "default":
                prompt = intelligent_question_filter_prompt.format(
                    question=question, file_name=file_name, passage=chunk
                )
                json_schema = {"Type_of_question": {"type": "string"}}
                response_json = self.llm.invoke(
                    prompt, json_schema, "IntelligentFilter", self.pipeline_id
                )
                try:
                    raw_answer = json.loads(response_json["reply"])
                except Exception as e:
                    print(
                        f"IntelligentFilter json decode error:{e}, with response_json:{response_json}"
                    )
                    raw_answer = {"Type_of_question": ""}
                filter_flag = raw_answer
            else:
                filter_flag = self.intelligent_question_filter.execute(
                    question, file_name, chunk
                )
            return question if filter_flag["Type_of_question"] == "Type_A" else None

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            filtered_questions = list(executor.map(filter_question, inputs))

        # Filter out any None values from the list of filtered questions
        questions = [q for q in filtered_questions if q is not None]
        docs = []
        for question in questions:
            doc = Document(
                content=question, meta={"file_name": file_name, "chunk": chunk}
            )
            docs.append(doc)
        # print("filtering intelligence with minimax done")
        return {"documents": docs}

    def to_dict(self) -> dict:
        return default_to_dict(self, service=self.service, pipeline_id=self.pipeline_id)

    @classmethod
    def from_dict(cls, data: dict) -> "IntelligentFilter":
        return default_from_dict(cls, data)



from abc import ABC, abstractmethod
import csv
from typing import Dict, List
from datasets import load_dataset
from creao.core.component.util import creao_component
from io import StringIO
import boto3
from urllib.parse import urlparse



class BaseDataComponent(ABC):
    def __init__(self, **kwargs):
        """
        Initialize the base data component, possibly with additional keyword arguments
        for future extensions.
        
        Args:
            **kwargs: Additional keyword arguments for future use.
        """
        self.global_dataset_map = kwargs.get('global_dataset_map', {})

    @abstractmethod
    def load_data(self) -> List[Dict[str, str]]:
        """
        Abstract method for loading data that must be implemented by subclasses.
        
        Returns:
            List[Dict[str, str]]: A list of dictionaries representing the loaded data.
        """
        pass

    def _remap_keys(self, dataset: List[Dict[str, str]]) -> List[Dict[str, str]]:
        """
        Apply the global dataset mapping to remap keys in the dataset.

        Args:
            dataset (List[Dict[str, str]]): The dataset to process.

        Returns:
            List[Dict[str, str]]: The dataset with keys remapped according to global_dataset_map.
        """
        remapped_data = []

        for item in dataset:
            # Create a copy of the item so that the original is not modified
            remapped_item = item.copy()
            # Iterate over each key in the global dataset map and remap it
            for key in self.global_dataset_map:
                if key in remapped_item:
                    value = remapped_item.pop(key)
                    mapped_key = self.global_dataset_map[key]
                    remapped_item[mapped_key] = value

            remapped_data.append(remapped_item)
        print("remapped_data:", remapped_data)

        return remapped_data

    def run(self, chained_input: List[Dict[str, str]] = None) -> List[Dict[str, str]]:
        """
        Load the data, apply global mappings, and return the processed data.

        Args:
            chained_input (List[Dict[str, str]], optional): Input data passed from the previous 
                                                            component in the pipeline.

        Returns:
            List[Dict[str, str]]: A list of dictionaries representing the processed dataset 
                                  with keys remapped according to the global dataset map.
        """
        # Load the data using the subclass-specific method
        dataset = self.load_data()

        # Apply global dataset mapping to the loaded data
        return self._remap_keys(dataset)
    

@creao_component
class HFDataComponent(BaseDataComponent):
    def __init__(self, hf_dataset_path: str, **kwargs):
        """
        Initialize the HFDataComponent by setting the path to the Hugging Face dataset.
        
        Args:
            hf_dataset_path (str): Path to the Hugging Face dataset (e.g., 'imdb').
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self.hf_path = hf_dataset_path

    def load_data(self) -> List[Dict[str, str]]:
        """
        Load data from the specified Hugging Face dataset and convert it to a list of dictionaries.
        
        Returns:
            List[Dict[str, str]]: A list of dictionaries representing the loaded dataset.
        """
        # Log the beginning of the dataset download process
        print(f"Downloading dataset from {self.hf_path}...")

        # Load the dataset split (defaulting to 'train')
        dataset = load_dataset(self.hf_path, split="train")
        
        # Convert the dataset to a list of dictionaries with string values
        return [{key: str(value) for key, value in example.items()} for example in dataset]


@creao_component
class CSVDataComponent(BaseDataComponent):
    def __init__(self, file_path: str, **kwargs):
        """
        Initialize the CSVDataComponent by setting the path to the S3 CSV file.

        Args:
            file_path (str): S3 path to the CSV file (e.g., 's3://bucket-name/path/to/file.csv').
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self.file_path = file_path

    def load_data(self) -> List[Dict[str, str]]:
        """
        Load data from the specified S3 CSV file and convert it to a list of dictionaries.
        
        Returns:
            List[Dict[str, str]]: A list of dictionaries representing the loaded CSV data.
        """
        s3_client = boto3.client('s3')
        s3_path = self.file_path
        # Parse the S3 bucket name and key from the S3 path
        bucket_name, key = self._parse_s3_url(s3_path)

        # Download the CSV file from S3
        print(f"Downloading CSV data from S3: {s3_path}...")
        response = s3_client.get_object(Bucket=bucket_name, Key=key)
        csv_content = response['Body'].read().decode('utf-8')

        # Read the CSV file and convert it to a list of dictionaries
        data_list = []
        reader = csv.DictReader(StringIO(csv_content))
        for row in reader:
            item = {key: str(value) for key, value in row.items()}
            print("item:", item)
            data_list.append(item)

        return data_list

    def _parse_s3_url(self, url: str) -> (str, str): # type: ignore
        """
        Parse the S3 URL to extract the bucket name and key.

        Args:
            url (str): The S3 URL (e.g., 'https://bucket-name.s3.region.amazonaws.com/key/to/file.csv').

        Returns:
            tuple: A tuple containing the bucket name and key.
        """
        parsed_url = urlparse(url)
        
        # Extract bucket name from the netloc (e.g., 'creaodev.s3.us-west-2.amazonaws.com')
        domain_parts = parsed_url.netloc.split('.')
        bucket_name = domain_parts[0]  # This assumes the bucket name is the first part
        
        # Extract key from the path (e.g., '/key/to/file.csv')
        key = parsed_url.path.lstrip('/')  # Strip leading slash from the path
        
        return bucket_name, key


from typing import Dict
from creao.component.generators.creaogenerator import CreaoGenerator
from haystack.document_stores.in_memory import InMemoryDocumentStore
from haystack import Document
from haystack.components.embedders import SentenceTransformersDocumentEmbedder
from haystack.components.embedders import SentenceTransformersTextEmbedder
from haystack.components.retrievers.in_memory import InMemoryEmbeddingRetriever
from haystack import Pipeline
from haystack.components.embedders import SentenceTransformersTextEmbedder
from haystack.components.builders import PromptBuilder
from haystack.components.generators import OpenAIGenerator
import gradio as gr


class RAG(object):
    def __init__(
        self,
        chunk: list[str],
        embedding_model_id: str = "BAAI/bge-base-en-v1.5",
        service="default",
    ) -> None:
        """
        :param chunk: The chunk to be stored in the vector database.
        """
        docs = [
            Document(content=item, meta={"file_path": "document.md"}) for item in chunk
        ]
        text_embedder = SentenceTransformersTextEmbedder(model=embedding_model_id)
        text_embedder.warm_up()
        doc_embedder = SentenceTransformersDocumentEmbedder(model=embedding_model_id)
        doc_embedder.warm_up()
        # store all chunk in a vector database
        document_store = InMemoryDocumentStore()
        docs_with_embeddings = doc_embedder.run(docs)
        document_store.write_documents(docs_with_embeddings["documents"])
        # create a retriever from the vector database
        retriever = InMemoryEmbeddingRetriever(document_store)
        # build the retriever pipeline with haystack
        template = """
        Given the following information, answer the question.

        Context:
        {% for document in documents %}
            {{ document.content }}
        {% endfor %}

        Question: {{question}}
        Answer:
        """
        prompt_builder = PromptBuilder(template=template)
        if service == "default":
            llm = CreaoGenerator()
        elif service == "openai":
            llm = OpenAIGenerator(model="gpt-4o-mini")
        basic_rag_pipeline = Pipeline()
        # Add components to your pipeline
        basic_rag_pipeline.add_component("text_embedder", text_embedder)
        basic_rag_pipeline.add_component("retriever", retriever)
        basic_rag_pipeline.add_component("prompt_builder", prompt_builder)
        basic_rag_pipeline.add_component("llm", llm)
        # Now, connect the components to each other
        basic_rag_pipeline.connect(
            "text_embedder.embedding", "retriever.query_embedding"
        )
        basic_rag_pipeline.connect("retriever", "prompt_builder.documents")
        basic_rag_pipeline.connect("prompt_builder", "llm")
        self.basic_rag_pipeline = basic_rag_pipeline

    def query_response(self, question: str) -> str:
        response = self.basic_rag_pipeline.run(
            {
                "text_embedder": {"text": question},
                "prompt_builder": {"question": question},
            },
            include_outputs_from={"retriever"},
        )
        return response

    def generate_gradio_interface(self, demo_examples: list[str] = []):
        def yes_man(message, history):
            response = self.basic_rag_pipeline.run(
                {
                    "text_embedder": {"text": message},
                    "prompt_builder": {"question": message},
                }
            )
            return response["llm"]["replies"][0]

        gr.ChatInterface(
            yes_man,
            chatbot=gr.Chatbot(height=300),
            textbox=gr.Textbox(placeholder="Ask me question", container=False, scale=7),
            title="Copilot for building Creao Component & Pipeline",
            description="Ask me question",
            theme="soft",
            examples=demo_examples,
            cache_examples=True,
            retry_btn=None,
            undo_btn="Delete Previous",
            clear_btn="Clear",
        ).launch(share=True)



import os
import requests
import networkx as nx
from itertools import product
from networkx.algorithms.dag import topological_sort
import yaml
import json
from creao.core.Endpoints import CreaoLLM, OpenAILLM
from creao.core.component.dedupe_component import Dedup, DedupeComponent
from creao.core.component.llm_component import LLMComponent
from creao.core.component.filter_component import FilterComponent
from creao.core.component.data_component import CSVDataComponent
from creao.core.component.util import class_registry, extract_json_schema, extract_jinja2_variables
from typing import List, Any, Dict
from pydantic import BaseModel, create_model

class Pipeline:
    def __init__(self):
        # Initialize an empty directed graph for the pipeline
        self.graph = nx.DiGraph()
        # Counter to assign unique IDs to each node
        self.node_id_counter = 0
        # Maps node names to node IDs for easy reference
        self.node_name_to_id = {}
        # Global variables shared across pipeline nodes
        self.global_variables = {}

    def add(self, name: str, class_instance):
        """
        Adds a node (component) to the pipeline.

        Args:
            name (str): The name of the node.
            class_instance: The instance of the class to be added as a node.
        """
        self.node_id_counter += 1
        # Add the node to the graph with a unique ID and store class instance
        self.graph.add_node(self.node_id_counter, name=name, data=class_instance)
        # Map the name to the node's ID for easy reference later
        self.node_name_to_id[name] = self.node_id_counter

    def connect(self, from_name: str, to_name: str):
        """
        Connects two nodes in the pipeline, forming an edge between them.

        Args:
            from_name (str): The name of the source node.
            to_name (str): The name of the target node.
        """
        from_node_id = self.node_name_to_id.get(from_name)
        to_node_id = self.node_name_to_id.get(to_name)
        if from_node_id is None or to_node_id is None:
            raise ValueError("One or both of the node names provided do not exist.")
        # Add a directed edge between two nodes in the graph
        self.graph.add_edge(from_node_id, to_node_id)

    def publish_pipeline_output(self, input_config, single_output):
        """
        Publishes the final pipeline output to a remote API.

        Args:
            input_config (dict): The input configuration used for the pipeline.
            single_output (dict): The final output to be published.
        """
        # API Gateway URL and necessary headers for publishing the output
        url = "https://drk3mkqa6b.execute-api.us-west-2.amazonaws.com/default/minimaxserver"
        api_key = os.environ["CREAO_API_KEY"]
        headers = {"Content-Type": "application/json", "x-api-key": api_key}

        # Prepare payload with pipeline and output data
        payload = {
            "single_output": single_output,
            "input_config": input_config,
            "pipeline_id": self.pipeline_id,
            "action": "pipeline_final_output",
        }

        # Send the request and handle the response
        response = requests.post(url, headers=headers, json=payload)
        print("publish_pipeline_output response:", response.content)

        # Attempt to return the response as JSON, handle any errors
        try:
            return response.json()
        except Exception as e:
            return {"error": str(e)}

    def run(self):
        """
        Runs the entire pipeline by executing the nodes in topological order.

        For each input from the data node, the output is passed to the next nodes
        in the pipeline. The final output is published via the publish function.
        """
        # Get the initial input data from the data node
        dict_list = self.run_datanode()
        for item in dict_list:
            print("input config:", item)
            # Run the pipeline for the current input
            output = self.run_single(item)
            print("single_output:", output)
            # Publish the output after the run
            self.publish_pipeline_output(item, output)

    def run_datanode(self):
        """
        Executes the first node in the pipeline (data node), and ensures all output values are lists.

        Returns:
            List[Dict[str, List[str]]]: The initial output from the data node.
        """
        # Ensure that the graph is a directed acyclic graph (DAG)
        if not nx.is_directed_acyclic_graph(self.graph):
            raise ValueError("The graph is not a directed acyclic graph (DAG).")

        # Get the nodes in topological order
        execution_order = list(topological_sort(self.graph))

        # Run the first node (data node) to get the initial input
        for node_id in execution_order:
            class_instance = self.graph.nodes[node_id]["data"]
            node_name = self.graph.nodes[node_id]["name"]
            # Execute the class instance to get the output list
            output_list = class_instance.run([], instance_name=node_name, validate_schema=False)

            # Ensure all values in the output are lists
            output_list_force_list_value = []
            for item in output_list:
                for key in item:
                    value = item[key]
                    if not isinstance(value, list):
                        item[key] = [value]
                output_list_force_list_value.append(item)
            break

        return output_list_force_list_value

    def run_single(self, input_dict: dict):
        """
        Executes the entire pipeline for a given input dictionary.

        Args:
            input_dict (Dict[str, List[str]]): The input configuration for the pipeline.

        Returns:
            dict: The final output after all nodes have processed the input.
        """
        output_store = {}

        # Ensure the graph is a directed acyclic graph (DAG)
        if not nx.is_directed_acyclic_graph(self.graph):
            raise ValueError("The graph is not a directed acyclic graph (DAG).")

        # Run nodes in topological order starting from the second node
        execution_order = list(topological_sort(self.graph))[1:]

        for node_id in execution_order:
            class_instance = self.graph.nodes[node_id]["data"]
            node_name = self.graph.nodes[node_id]["name"]
            # Get predecessors of the current node
            predecessors = list(self.graph.predecessors(node_id))

            # Collect the output from predecessor nodes
            predecessors_list = [output_store[self.graph.nodes[item]["data"]] for item in predecessors if self.graph.nodes[item]["data"] in output_store]

            print("predecessors_list:", predecessors_list)

            # If any predecessor output is None, return an empty list
            if None in predecessors_list:
                return []

            # Generate all possible input combinations
            combinations = list(product(*predecessors_list))
            chained_input = [{k: v for d in item for k, v in d.items()} for item in combinations]

            if not chained_input:
                chained_input = [input_dict]

            # Run the class instance with the combined input
            combined_global_variables = self.global_variables.copy()
            combined_global_variables.update(input_dict)
            output = class_instance.run(chained_input, global_variables=combined_global_variables, instance_name=node_name, validate_schema=False)

            # Store the output for downstream nodes
            output_store[class_instance] = output

        return output

    def to_dict(self):
        """
        Converts the pipeline structure into a dictionary format for serialization.

        Returns:
            dict: The pipeline represented as a dictionary.
        """
        pipeline_dict = {"nodes": [], "connections": []}

        # Add nodes and their configurations to the dictionary
        for node_id, node_data in self.graph.nodes(data=True):
            class_instance = node_data["data"]
            class_dict = class_instance.__dict__.copy()

            # Remove sensitive data like LLM instance from serialization
            class_dict.pop("llm", None)

            node_dict = {
                "name": node_data["name"],
                "class": class_instance.__class__.__name__,
                "params": class_dict,
            }
            pipeline_dict["nodes"].append(node_dict)

        # Add connections between nodes to the dictionary
        for from_node, to_node in self.graph.edges():
            from_name = self.graph.nodes[from_node]["name"]
            to_name = self.graph.nodes[to_node]["name"]
            pipeline_dict["connections"].append({"from": from_name, "to": to_name})

        return pipeline_dict

    def save_to_yaml(self, file_path):
        """
        Saves the pipeline structure to a YAML file.

        Args:
            file_path (str): The path to save the YAML file.
        """
        pipeline_dict = self.to_dict()
        with open(file_path, "w") as yaml_file:
            yaml.dump(pipeline_dict, yaml_file)

    @staticmethod
    def from_dict(pipeline_dict, dataset_children_map, global_variables):
        """
        Reconstructs a Pipeline instance from a dictionary representation.

        Args:
            pipeline_dict (dict): The dictionary representation of the pipeline.
            dataset_children_map (dict): A map of dataset relationships.
            global_variables (dict): Global variables to be used in the pipeline.

        Returns:
            Pipeline: The reconstructed pipeline instance.
        """
        pipeline = Pipeline()
        pipeline.pipeline_id = pipeline_dict.get("pipeline_id", "pipeline_id_default")
        pipeline.global_variables = global_variables
        pipeline.dataset_children_map = dataset_children_map

        # Reconstruct nodes
        for node in pipeline_dict["nodes"]:
            class_name = node["class"]
            class_instance = class_registry[class_name](**node["params"])
            if class_name == "CSVDataComponent":
                class_instance.global_dataset_map = pipeline_dict.get("global_dataset_map", {})
            pipeline.add(node["name"], class_instance)

        # Reconstruct connections
        for connection in pipeline_dict["connections"]:
            pipeline.connect(connection["from"], connection["to"])

        # Reinitialize necessary components for certain classes
        for node_id in pipeline.graph.nodes:
            class_instance = pipeline.graph.nodes[node_id]["data"]
            if isinstance(class_instance, DedupeComponent):
                class_instance.llm = Dedup()
            elif isinstance(class_instance, LLMComponent):
                if class_instance.service == "default":
                    class_instance.llm = CreaoLLM(bot_name="assistant", bot_content="assistant")
                elif class_instance.service == "openai":
                    class_instance.llm = OpenAILLM()

        return pipeline

    @staticmethod
    def create_pydantic_model(fields: List[Dict[str, Any]]) -> BaseModel:
        """
        Create a Pydantic model dynamically from a list of dictionaries representing fields.

        Args:
        - fields (List[Dict[str, Any]]): A list where each dict contains 'name' and 'type' keys.

        Returns:
        - A dynamically created Pydantic model class.
        """
        new_json_list = []
        for item in fields:
            if item["type"] == "list[str]":
                item["type"] = List[str]
            elif item["type"] == "str":
                item["type"] = str
            new_json_list.append(item)
        fields = new_json_list
        field_definitions = {field['name']: (field['type'], ...) for field in fields}
        return create_model('JsonBase', **field_definitions)
    

    @staticmethod
    def load_from_rf_json(file_path, rf_dict=None, global_variables=None):
        """
        Loads a pipeline from a JSON file created in a specific format.

        Args:
            file_path (str): The path to the JSON file.
            rf_dict (dict): The dictionary representation of the pipeline (optional).
            global_variables (dict): Global variables for the pipeline (optional).

        Returns:
            Pipeline: The reconstructed pipeline.
        """
        if rf_dict is None:
            with open(file_path, "r") as rf_file:
                rf_dict = json.load(rf_file)

        nodes = []
        id_to_name_dict = {}

        # Reconstruct nodes from the JSON file
        for node in rf_dict["nodes"]:
            id = node["id"]
            temp_dict = {}
            params = {}
            node_data = node["data"]
            output_schema = node_data.get("output_schema", [])

            temp_dict["name"] = node_data["label"]

            # Assign class based on the node type
            if "type" not in node:
                temp_dict["class"] = "LLMComponent"
                json_schema = extract_json_schema(Pipeline.create_pydantic_model(output_schema))
                params["json_schema"] = json_schema
            elif node["type"] == "dataNode":
                temp_dict["class"] = "CSVDataComponent"
            elif node["type"] == "filterNode":
                temp_dict["class"] = "FilterComponent"
            elif node["type"] == "dedupeNode":
                temp_dict["class"] = "DedupeComponent"

            # Add additional parameters to the node
            for key in node_data:
                if key not in ("output_schema", "label"):
                    params[key] = node_data[key]

            params["component_name"] = node_data["label"]
            temp_dict["params"] = params

            prompt_template = node_data.get("prompt_template", "")
            id_to_name_dict[id] = {
                "label": node_data["label"],
                "class": temp_dict["class"],
                "prompt_template": prompt_template,
            }
            nodes.append(temp_dict)

        # Reconstruct connections and dataset children map
        connections = []
        dataset_children_map = {}
        for edge in rf_dict["edges"]:
            from_node = id_to_name_dict[edge["source"]]
            to_node = id_to_name_dict[edge["target"]]

            # Handle special case for CSVDataComponent nodes
            if from_node["class"] == "CSVDataComponent":
                dataset_children_map.setdefault(from_node["label"], {})[to_node["label"]] = extract_jinja2_variables(to_node["prompt_template"])

            connections.append({"from": from_node["label"], "to": to_node["label"]})

        # Final JSON to be processed back into a pipeline
        final_json = {
            "nodes": nodes,
            "connections": connections,
            "pipeline_id": rf_dict["pipeline_id"],
            "global_dataset_map": rf_dict["global_dataset_map"],
        }
        print("final_json:", final_json)
        return Pipeline.from_dict(final_json, dataset_children_map, global_variables)




from typing import Dict, List, Union
from creao.core.component.util import creao_component

@creao_component  # This decorator registers the class as a component in a pipeline framework
class FilterComponent:
    """
    FilterComponent is responsible for filtering data based on a specific condition applied to a designated column.
    The component can perform two types of filtering:
    1. Exact match filtering for strings or direct equality comparison.
    2. Numeric condition filtering for values that need to meet a specified numeric condition (e.g., >, <, >=).

    Attributes:
    - filtered_column (str): The name of the column on which the filter is applied.
    - condition_type (str): The type of filtering condition. It can be either "exact_match" or "numeric_condition".
    - condition_value (Union[str, int, float]): The value or condition used for filtering. It can be a string (for exact match)
      or a numeric condition (e.g., "> 5", "== 10").
    - pipeline_id (str): Identifier for the pipeline this component belongs to. Defaults to "pipeline_id_default".
    - component_name (str): Name of the component, default is "FilterComponent".
    - **kwargs: Additional arguments that can be passed for customization.

    Methods:
    - run(chained_input: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
      This method performs the filtering on the input data (a list of dictionaries). It processes each dictionary (or row)
      and filters it based on the specified conditions. It returns the filtered data as a list of dictionaries.

      - chained_input: The input data, which is expected to be a list of dictionaries where each dictionary represents a row or record.
      - filtered_output: The filtered data after applying the conditions.

      The method supports:
      - "exact_match" condition: Filters based on string equality or direct comparison to a specific value.
      - "numeric_condition" condition: Filters based on a numerical comparison using operators like >, <, >=, etc.

      Example usage:
      - If the condition_type is "exact_match" and condition_value is "apple", it will filter and return only the rows where
        the specified column has the value "apple".
      - If the condition_type is "numeric_condition" and condition_value is "> 5", it will filter and return rows where the
        numeric value in the specified column is greater than 5.

      If the column value cannot be found in the row or cannot be converted to a number for numeric conditions, the row is skipped.

    Example:
    component = FilterComponent(filtered_column="price", condition_type="numeric_condition", condition_value="> 50")
    filtered_data = component.run(input_data)
    """

    def __init__(
        self,
        filtered_column: str,
        condition_type: str = "exact_match",  # or "numeric_condition"
        condition_value: Union[str, int, float] = None,
        pipeline_id: str = "pipeline_id_default",
        component_name: str = "FilterComponent",
        **kwargs,
    ):
        self.filtered_column = filtered_column  # Column to apply the filter on
        self.condition_type = condition_type  # Type of condition, either "exact_match" or "numeric_condition"
        self.condition_value = condition_value  # Value or condition to filter against
        self.pipeline_id = pipeline_id  # ID of the pipeline this component belongs to
        self.component_name = component_name  # Name of the component

    def run(
        self, chained_input: List[Dict[str, List[str]]]
    ) -> List[Dict[str, List[str]]]:
        for single_chained_input in chained_input:
            assert (
                self.filtered_column in single_chained_input
            ), f"Column '{self.filtered_column}' not found in the input data."
            filtered_output = []
            column_values = single_chained_input.get(
                self.filtered_column
            )  # Retrieve the value from the specified column
            # Handle different condition types
            for column_value in column_values:
                if self.condition_type == "exact_match":
                    if column_value == self.condition_value:  # Check for exact match
                        filtered_output.append(
                            column_value
                        )  # Add the row to the output if it matches

                elif self.condition_type == "numeric_condition":
                    try:
                        numeric_value = float(
                            column_value
                        )  # Convert the column value to a numeric value
                        if eval(
                            f"{numeric_value} {self.condition_value}"
                        ):  # Evaluate the numeric condition
                            filtered_output.append(
                                column_value
                            )  # Add the row to the output if it meets the condition
                    except ValueError:
                        continue  # Skip if the value can't be converted to a number
            single_chained_input[self.filtered_column] = filtered_output
        return chained_input  # Return the filtered data


from itertools import product
from typing import Any, Dict, List

from jinja2 import Environment, meta

# A global registry to store all registered classes
class_registry = {}


def creao_component(cls):
    """
    Registers a class in the global registry and wraps its `run` method
    to handle global variables.

    Args:
        cls (type): The class to be registered and wrapped.

    Returns:
        type: The original class with its `run` method wrapped.
    """
    # Retrieve the original `run` method of the class
    original_run_method = getattr(cls, "run", None)

    # Ensure the class has a `run` method
    if original_run_method is None:
        raise ValueError(f"Class {cls.__name__} must have a 'run' method.")

    def wrapped_run(self, chained_input: List[Dict[str, List[str]]], *args, **kwargs):
        """
        Wrapper for the original `run` method that processes global variables
        and updates the chained input accordingly.

        Args:
            chained_input (List[Dict[str, List[str]]]): The input data that may be processed
                                                        and passed to the `run` method.
            *args: Additional positional arguments for the `run` method.
            **kwargs: Additional keyword arguments for the `run` method,
                      including `global_variables` which are processed separately.

        Returns:
            Any: The output from the original `run` method.
        """
        # Extract and process global variables from kwargs
        global_variables = kwargs.pop("global_variables", {})
        processed_global_variables = {}

        # Convert global variables into a consistent list format
        for key, value in global_variables.items():
            processed_global_variables[key] = (
                value if isinstance(value, list) else [value]
            )

        # Update each item in the chained input with the processed global variables
        for single_chained_input in chained_input:
            single_chained_input.update(processed_global_variables)

        # Call the original `run` method with the modified input
        output = original_run_method(self, chained_input)
        return output

    # Replace the original `run` method of the class with the wrapped version
    setattr(cls, "run", wrapped_run)

    # Register the class in the global class registry
    class_registry[cls.__name__] = cls

    # Return the original class with the wrapped `run` method
    return cls


def generate_combinations_from_dict_list(dict_list):
    """
    Generate a list of dictionaries representing all possible combinations of values
    from a list of dictionaries with identical keys.

    Args:
        dict_list (list of dict): A list of dictionaries where each dictionary has the same keys.

    Returns:
        list of dict: A list of dictionaries, each representing a unique combination of the values
                      from the input dictionaries.
    """
    if len(dict_list) == 0:
        return []

    # Initialize lists to store keys and values from the input dictionaries
    keys = []
    values = []

    # Extract keys and corresponding values from each dictionary in the input list
    for d in dict_list:
        for k, v in d.items():
            keys.append(k)
            values.append(v)

    # Generate the Cartesian product of the lists of values
    combinations = product(*values)

    # Construct a list of dictionaries, each representing a unique combination
    result = [dict(zip(keys, combo)) for combo in combinations]

    return result


def extract_json_schema(class_instance) -> Dict[str, Any]:
    """
    Extract and modify the JSON schema of a given class instance.

    Args:
        class_instance: An instance of a class that has a `.schema()` method which returns a JSON schema.

    Returns:
        Dict[str, Any]: A dictionary representing the modified JSON schema, with 'title' fields removed.
    """
    # Generate the JSON schema for the class instance
    json_schema = class_instance.schema()

    # Extract the 'properties' part of the schema, which contains the details of each field
    output_dict = json_schema["properties"]

    # Iterate over each field in the class instance
    for key in class_instance.__fields__:
        # Remove the 'title' attribute from each field's schema (if present)
        output_dict[key].pop("title")

    # Return the modified schema as a dictionary
    return output_dict


def extract_jinja2_variables(template_str: str) -> List[str]:
    env = Environment()
    # Parse the template
    parsed_content = env.parse(template_str)
    variables = meta.find_undeclared_variables(parsed_content)
    return list(variables)



from setuptools import setup, find_packages

setup(
    name="creao",
    version="0.3.12",
    packages=find_packages(),
    install_requires=[
        "networkx==3.3",
        "openai==1.40.6",
        "datasets==2.21.0",
    ],
    include_package_data=True,
    description="A description of your project",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://creao@dev.azure.com/creao/creao/_git/creao",
    author="creaoai",
    author_email="dev@creao.ai",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import json
from typing import Dict, List, Tuple

from jinja2 import Template
from creao.core.Endpoints import CreaoLLM, OpenAILLM
from creao.core.component.util import (
    extract_jinja2_variables,
    creao_component,
    generate_combinations_from_dict_list,
    extract_json_schema,
)


@creao_component
class LLMComponent:
    def __init__(
        self,
        prompt_template: str = None,
        class_instance=None,
        service: str = "default",
        pipeline_id: str = "pipeline_id_default",
        component_name: str = "default",
        **kwargs,
    ):
        """
        Initializes the LLMComponent with necessary configurations.

        Args:
            prompt_template (str): Template string for generating prompts using Jinja2.
            class_instance: An instance of a class used to extract a JSON schema (optional).
            service (str): Specifies the LLM service to use, default is "default".
            pipeline_id (str): Identifier for the pipeline, default is "pipeline_id_default".
            component_name (str): Name of the component, default is "default".
            **kwargs: Additional keyword arguments, such as `json_schema` and `type`.
        """
        self.prompt_template = (
            prompt_template  # Template to be used for generating prompts
        )
        self.component_name = (
            component_name  # Name of the component for logging/tracking
        )
        self.pipeline_id = (
            pipeline_id  # Identifier for the pipeline this component belongs to
        )
        self.service = service  # The LLM service type (e.g., "default", "openai")

        # Extract JSON schema from the provided class instance or use the one in kwargs
        self.json_schema = self._extract_json_schema(class_instance, kwargs)

        # Initialize the appropriate LLM instance based on the selected service
        self.llm = self._initialize_llm()

    def run(
        self, chained_input: List[Dict[str, List[str]]]
    ) -> List[Dict[str, List[str]]]:
        """
        Processes the input data through the LLM based on the prompt template.

        Args:
            chained_input (List[Dict[str, List[str]]]): List of dictionaries containing input data.

        Returns:
            List[Dict[str, List[str]]]: Processed output after invoking the LLM.
        """
        # Extract variables from the Jinja2 template to be used in rendering the prompt
        extracted_vars = self._extract_jinja2_variables()

        res_list = []  # Initialize a list to store results for each input
        for single_chain_input in chained_input:
            # Prepare Jinja2 inputs by mapping extracted variables to the chain input
            jinja_inputs = self._prepare_jinja_inputs(
                extracted_vars, single_chain_input
            )

            # Generate a list of prompts by rendering the template with Jinja2 inputs
            prompts = self._generate_prompts(jinja_inputs)

            for prompt, input_dict in prompts:
                # Invoke the LLM with the rendered prompt and handle the response
                response_json = self.invoke_llm(prompt)
                print("response_json:", response_json)

                if response_json is None:
                    return None  # Exit early if there's an error invoking the LLM

                # Process the LLM's response and prepare the final output format
                processed_response = self.process_llm_response(
                    response_json, input_dict
                )
                res_list.append(
                    processed_response
                )  # Append the processed response to the result list

        return res_list  # Return the list of processed results

    def _extract_json_schema(self, class_instance, kwargs) -> dict:
        """
        Extracts the JSON schema from the class instance or uses the provided schema.

        Args:
            class_instance: The class instance from which to extract the JSON schema.
            kwargs: Dictionary containing the keyword arguments, including `json_schema`.

        Returns:
            dict: The extracted or provided JSON schema.
        """
        json_schema = kwargs.pop(
            "json_schema", None
        )  # Attempt to get the JSON schema from kwargs
        if json_schema is None and class_instance is not None:
            # If no schema is provided, try to extract it from the class instance
            return extract_json_schema(class_instance)
        return json_schema  # Return the schema from kwargs or None if not found

    def _initialize_llm(self):
        """
        Initializes the LLM based on the selected service.

        Returns:
            The initialized LLM instance.
        """
        # Initialize the LLM based on the service type
        if self.service == "default":
            return CreaoLLM(bot_name="assistant", bot_content="assistant")
        elif self.service == "openai":
            return OpenAILLM()
        return None  # Return None if an unsupported service is selected

    def _extract_jinja2_variables(self) -> List[str]:
        """
        Extracts variables from the Jinja2 template.

        Returns:
            List[str]: A list of variables extracted from the Jinja2 template.
        """
        # Use the Pipeline utility to extract variables from the prompt template
        return extract_jinja2_variables(self.prompt_template)

    def _prepare_jinja_inputs(
        self, extracted_vars: List[str], single_chain_input: Dict[str, List[str]]
    ) -> List[Dict[str, str]]:
        """
        Prepares Jinja2 inputs by mapping extracted variables to the chain input.

        Args:
            extracted_vars (List[str]): List of variables extracted from the template.
            single_chain_input (Dict[str, List[str]]): Single input dictionary from the chained input.

        Returns:
            List[Dict[str, str]]: A list of dictionaries prepared for Jinja2 rendering.
        """
        # Create a list of dictionaries where each variable is mapped to its corresponding value
        jinja_var_list = [{var: single_chain_input[var]} for var in extracted_vars]
        return generate_combinations_from_dict_list(
            jinja_var_list
        )  # Convert the list of dictionaries into the required format

    def _generate_prompts(
        self, jinja_inputs: List[Dict[str, str]]
    ) -> List[Tuple[str, Dict[str, str]]]:
        """
        Generates a list of prompts by rendering the template with Jinja2 inputs.

        Args:
            jinja_inputs (List[Dict[str, str]]): A list of dictionaries prepared for Jinja2 rendering.

        Returns:
            List[Tuple[str, Dict[str, str]]]: A list of tuples containing the rendered prompt and the corresponding input dictionary.
        """
        # Initialize the Jinja2 template with the prompt template string
        prompt_template = Template(self.prompt_template)
        # Render the template with each set of inputs and return the list of prompts with input mappings
        return [(prompt_template.render(item), item) for item in jinja_inputs]

    def invoke_llm(self, prompt: str) -> dict:
        """
        Invokes the appropriate LLM based on the service type.

        Args:
            prompt (str): The prompt to be sent to the LLM.

        Returns:
            dict: The LLM's response as a dictionary, or None if an error occurs.
        """
        try:
            if self.service == "default":
                # Invoke the default LLM with the prompt, schema, component name, and pipeline ID
                response_json = self.llm.invoke(
                    prompt,
                    self.json_schema or "",
                    self.component_name,
                    self.pipeline_id,
                )
                if self.json_schema:
                    # Parse the response as JSON if a schema is provided
                    return json.loads(response_json["reply"])
                return {
                    "reply": response_json["reply"]
                }  # Return the plain reply in a dict with key as reply if no schema is used
            elif self.service == "openai":
                # For OpenAI service, simply parse the response as JSON
                return json.loads(self.llm.invoke(prompt))
        except Exception as e:
            # Log any exceptions that occur during LLM invocation
            print(f"Error invoking LLM: {e}")
            return None  # Return None if an error occurs

    def process_llm_response(self, raw_answer: dict, input_dict: dict) -> dict:
        """
        Processes the LLM's raw response into a consistent list-based dictionary.

        Args:
            raw_answer (dict): The raw response from the LLM.
            input_dict (dict): The original input dictionary.

        Returns:
            dict: Processed response with each value in list format.
        """
        processed_response = {}

        if self.service == "openai":
            # For OpenAI service, wrap the raw answer in a list under the "reply" key
            processed_response["reply"] = [raw_answer]
        else:
            # For the default service, ensure all values in the response are lists
            for key, value in raw_answer.items():
                processed_response[key] = value if isinstance(value, list) else [value]

        # Update the processed response with the original input values, converting them to lists
        input_dict_value_list = {key: [input_dict[key]] for key in input_dict}
        processed_response.update(input_dict_value_list)

        return processed_response  # Return the final processed response



# import all class within mappers
from .map_interest_to_question_type import MapInterestToQuestionType
from .map_question_type_to_question import MapQuestionTypeToQuestion
from .rewrite_question import (
    RewriteQuestion,
    PersonaToWritingStyle,
    RewriteQuestionByPersona,
)
from .text_to_persona import TextToPersona



from collections import defaultdict
from creao.core.component.util import creao_component
import concurrent.futures
from typing import Dict, List
import numpy as np
from creao.core.Endpoints import CreaoLLM
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics.pairwise import cosine_distances


class Dedup:
    def __init__(self):
        self.embedding_model = CreaoLLM()

    def get_embedding_vector(self, text):
        return self.embedding_model.invoke(text, "", component_id="embed")

    def list2vec(self, text_list, num_workers=100):
        def process_text(text):
            vector = self.get_embedding_vector(text)
            return text, vector

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
            results = list(executor.map(process_text, text_list))

        texts, embeddings = zip(*results)
        embeddings = np.array(embeddings)

        return list(texts), embeddings

    def clustering(self, embeddings, threshold=0.075):
        cosine_dist_matrix = cosine_distances(embeddings)

        agg_clustering = AgglomerativeClustering(
            n_clusters=None,
            metric="precomputed",
            linkage="complete",
            distance_threshold=threshold,
        )
        labels = agg_clustering.fit_predict(cosine_dist_matrix)

        return labels

    def execute(self, text_list: List[str]) -> List[str]:
        if len(text_list) < 2:
            return text_list
        try:
            texts, embeddings = self.list2vec(text_list)
        except Exception as e:
            print(f"dedupe list2vec with error: {e}")
            return text_list

        labels = self.clustering(embeddings)
        assert len(texts) == len(labels)
        # only keep the first text with duplicated label, and return the list[{text, drop}], where drop is a boolean
        exist_label = set()
        res = []
        for i in range(len(texts)):
            label = labels[i]
            if label in exist_label:
                continue
            else:
                exist_label.add(label)
                res.append(texts[i])
        return res
    
@creao_component
class DedupeComponent:
    def __init__(
        self,
        dedup_column: str,
        pipeline_id: str = "pipeline_id_default",
        component_name: str = "default",
        **kwargs,
    ):
        self.pipeline_id = pipeline_id
        self.dedup_column = dedup_column
        self.component_name = component_name
        self.llm = Dedup()

    def convert_to_dict_of_lists(
        self, data: List[Dict[str, str]]
    ) -> Dict[str, List[str]]:
        result = defaultdict(list)

        for entry in data:
            for key, value in entry.items():
                result[key].append(value)
        return result

    def run(
        self, chained_input: List[Dict[str, List[str]]]
    ) -> List[Dict[str, List[str]]]:
        """
        Deduplicate the input texts.

        Args:
            chained_input (List[Dict[str, str]]): Input from previous component.
        Returns:
            List[Dict[str, str]]: Deduplicated texts.
        """
        for single_chained_input in chained_input:
            assert (
                self.dedup_column in single_chained_input
            ), f"Column '{self.dedup_column}' not found in the input data."
            for key in single_chained_input:
                if key == self.dedup_column:
                    texts = single_chained_input[key]
                    dedupe_res = self.llm.execute(texts)
                    single_chained_input[key] = dedupe_res
        return chained_input
